@@ -3,11 +3,13 @@ package librpki
 import (
 	"bytes"
 	"crypto/x509"
+	"crypto/x509/pkix"
 	"encoding/asn1"
 	"encoding/hex"
 	"errors"
 	"fmt"
 	"net"
+	"sort"
 	"time"
 )
 
@@ -22,10 +24,20 @@ var (
 	AutonomousSysIdsV2 = asn1.ObjectIdentifier{1, 3, 6, 1, 5, 5, 7, 1, 29}
 	IpAddrAndASIdent   = asn1.ObjectIdentifier{1, 3, 6, 1, 5, 5, 7, 1, 30}
 
-	SubjectInfoAccess = asn1.ObjectIdentifier{1, 3, 6, 1, 5, 5, 7, 1, 11}
+	CertPolicy         = asn1.ObjectIdentifier{2, 5, 29, 32}
+	ResourceCertPolicy = asn1.ObjectIdentifier{1, 3, 6, 1, 5, 5, 7, 14, 2}
+	CPS                = asn1.ObjectIdentifier{1, 3, 6, 1, 5, 5, 7, 2, 1}
+
+	SubjectInfoAccess   = asn1.ObjectIdentifier{1, 3, 6, 1, 5, 5, 7, 1, 11}
+	AuthorityInfoAccess = asn1.ObjectIdentifier{1, 3, 6, 1, 5, 5, 7, 1, 1}
+	CAIssuer            = asn1.ObjectIdentifier{1, 3, 6, 1, 5, 5, 7, 48, 2}
+	SignedObject        = asn1.ObjectIdentifier{1, 3, 6, 1, 5, 5, 7, 48, 11}
 
 	SubjectKeyIdentifier   = asn1.ObjectIdentifier{2, 5, 29, 14}
 	AuthorityKeyIdentifier = asn1.ObjectIdentifier{2, 5, 29, 35}
+
+	CertRepository = asn1.ObjectIdentifier{1, 3, 6, 1, 5, 5, 7, 48, 5}
+	CertRRDP       = asn1.ObjectIdentifier{1, 3, 6, 1, 5, 5, 7, 48, 13}
 )
 
 type IPNet struct {
@@ -54,6 +66,10 @@ func (ipn *IPNet) GetRange() (net.IP, net.IP, bool) {
 	return min, max, false
 }
 
+func (ipn *IPNet) ASN1() ([]byte, error) {
+	return asn1.Marshal(IPNetToBitString(*ipn.IPNet))
+}
+
 type IPAddressRange struct {
 	Min net.IP
 	Max net.IP
@@ -85,6 +101,42 @@ func (ipr *IPAddressRange) GetRange() (net.IP, net.IP, bool) {
 	return ipr.Min, ipr.Max, false
 }
 
+func IPToBitString(ip net.IP) asn1.BitString {
+	blen := 32
+	bitsplit := 12
+	if ip.To4() == nil {
+		blen = 128
+		bitsplit = 0
+	}
+	return asn1.BitString{
+		Bytes:     []byte(ip)[bitsplit:],
+		BitLength: blen,
+	}
+}
+
+func IPNetToBitString(ipnet net.IPNet) asn1.BitString {
+	size, _ := ipnet.Mask.Size()
+	sizeBytes := size / 8
+	if size%8 > 0 {
+		sizeBytes++
+	}
+	prefixBytes := make([]byte, sizeBytes)
+	for i := 0; i < len(ipnet.IP) && i < sizeBytes; i++ {
+		prefixBytes[i] = ipnet.IP[i] & ipnet.Mask[i]
+	}
+
+	return asn1.BitString{
+		Bytes:     prefixBytes,
+		BitLength: size,
+	}
+}
+
+func (ipr *IPAddressRange) ASN1() ([]byte, error) {
+	return asn1.Marshal([]asn1.BitString{IPToBitString(ipr.Min), IPToBitString(ipr.Max)})
+}
+
+// Add IP address type (just bit string)
+
 type IPAddressNull struct {
 	Family uint8
 }
@@ -110,17 +162,25 @@ func (ipan *IPAddressNull) GetRange() (net.IP, net.IP, bool) {
 	return nil, nil, true
 }
 
+func (ipan *IPAddressNull) ASN1() ([]byte, error) {
+	return asn1.Marshal(asn1.NullRawValue)
+}
+
 type IPCertificateInformation interface {
 	GetRange() (net.IP, net.IP, bool)
 	IsIPInRange(net.IP) (bool, bool)
 	String() string
 	GetAfi() uint8
+
+	ASN1() ([]byte, error)
 }
 
 type ASNCertificateInformation interface {
 	GetRange() (int, int, bool)
 	IsASNInRange(int) (bool, bool)
 	String() string
+
+	ASN1() ([]byte, error)
 }
 
 type ASNRange struct {
@@ -140,6 +200,10 @@ func (ar *ASNRange) GetRange() (int, int, bool) {
 	return ar.Min, ar.Max, false
 }
 
+func (ar *ASNRange) ASN1() ([]byte, error) {
+	return asn1.Marshal([]int{ar.Min, ar.Max})
+}
+
 type ASN struct {
 	ASN int
 }
@@ -156,6 +220,10 @@ func (a *ASN) GetRange() (int, int, bool) {
 	return a.ASN, a.ASN, false
 }
 
+func (a *ASN) ASN1() ([]byte, error) {
+	return asn1.Marshal(a.ASN)
+}
+
 type ASNull struct {
 }
 
@@ -171,6 +239,10 @@ func (an *ASNull) GetRange() (int, int, bool) {
 	return 0, 0, true
 }
 
+func (an *ASNull) ASN1() ([]byte, error) {
+	return asn1.Marshal(asn1.NullRawValue)
+}
+
 func DecodeIP(addrfamily []byte, addr asn1.BitString) (*net.IPNet, error) {
 	if len(addrfamily) >= 2 && (addrfamily[1] == 1 || addrfamily[1] == 2) {
 		size := 4
@@ -350,7 +422,7 @@ func DecodeASN(data []byte) ([]ASNCertificateInformation, []ASNCertificateInform
 	return asnsnum, asnsrdi, nil
 }
 
-type RPKI_Certificate struct {
+type RPKICertificate struct {
 	SubjectInformationAccess []SIA
 	IPAddresses              []IPCertificateInformation
 	ASNums                   []ASNCertificateInformation
@@ -362,7 +434,7 @@ type RPKI_Certificate struct {
 	//AuthorityKeyIdentifier []byte
 }
 
-func (cert *RPKI_Certificate) IsIPRangeInCertificate(min net.IP, max net.IP) (bool, bool) {
+func (cert *RPKICertificate) IsIPRangeInCertificate(min net.IP, max net.IP) (bool, bool) {
 	for _, ip := range cert.IPAddresses {
 		minIn, checkParentMin := ip.IsIPInRange(min)
 		maxIn, checkParentMax := ip.IsIPInRange(max)
@@ -377,7 +449,7 @@ func (cert *RPKI_Certificate) IsIPRangeInCertificate(min net.IP, max net.IP) (bo
 	return false, false
 }
 
-func (cert *RPKI_Certificate) IsASRangeInCertificate(min int, max int) (bool, bool) {
+func (cert *RPKICertificate) IsASRangeInCertificate(min int, max int) (bool, bool) {
 	for _, asn := range cert.ASNums {
 		minIn, checkParentMin := asn.IsASNInRange(min)
 		maxIn, checkParentMax := asn.IsASNInRange(max)
@@ -392,7 +464,7 @@ func (cert *RPKI_Certificate) IsASRangeInCertificate(min int, max int) (bool, bo
 }
 
 // https://tools.ietf.org/html/rfc6487#section-7.2
-func ValidateIPCertificateList(list []IPCertificateInformation, parent *RPKI_Certificate) ([]IPCertificateInformation, []IPCertificateInformation, []IPCertificateInformation) {
+func ValidateIPCertificateList(list []IPCertificateInformation, parent *RPKICertificate) ([]IPCertificateInformation, []IPCertificateInformation, []IPCertificateInformation) {
 	valids := make([]IPCertificateInformation, 0)
 	invalids := make([]IPCertificateInformation, 0)
 	checkParents := make([]IPCertificateInformation, 0)
@@ -414,11 +486,11 @@ func ValidateIPCertificateList(list []IPCertificateInformation, parent *RPKI_Cer
 	return valids, invalids, checkParents
 }
 
-func (cert *RPKI_Certificate) ValidateIPCertificate(parent *RPKI_Certificate) ([]IPCertificateInformation, []IPCertificateInformation, []IPCertificateInformation) {
+func (cert *RPKICertificate) ValidateIPCertificate(parent *RPKICertificate) ([]IPCertificateInformation, []IPCertificateInformation, []IPCertificateInformation) {
 	return ValidateIPCertificateList(cert.IPAddresses, parent)
 }
 
-func ValidateASNCertificateList(list []ASNCertificateInformation, parent *RPKI_Certificate) ([]ASNCertificateInformation, []ASNCertificateInformation, []ASNCertificateInformation) {
+func ValidateASNCertificateList(list []ASNCertificateInformation, parent *RPKICertificate) ([]ASNCertificateInformation, []ASNCertificateInformation, []ASNCertificateInformation) {
 	valids := make([]ASNCertificateInformation, 0)
 	invalids := make([]ASNCertificateInformation, 0)
 	checkParents := make([]ASNCertificateInformation, 0)
@@ -440,11 +512,11 @@ func ValidateASNCertificateList(list []ASNCertificateInformation, parent *RPKI_C
 	return valids, invalids, checkParents
 }
 
-func (cert *RPKI_Certificate) ValidateASNCertificate(parent *RPKI_Certificate) ([]ASNCertificateInformation, []ASNCertificateInformation, []ASNCertificateInformation) {
+func (cert *RPKICertificate) ValidateASNCertificate(parent *RPKICertificate) ([]ASNCertificateInformation, []ASNCertificateInformation, []ASNCertificateInformation) {
 	return ValidateASNCertificateList(cert.ASNums, parent)
 }
 
-func (cert *RPKI_Certificate) Validate(parent *RPKI_Certificate) error {
+func (cert *RPKICertificate) Validate(parent *RPKICertificate) error {
 	if cert.Certificate == nil {
 		return errors.New("No certificate found")
 	}
@@ -458,7 +530,7 @@ func (cert *RPKI_Certificate) Validate(parent *RPKI_Certificate) error {
 	return nil
 }
 
-func (cert *RPKI_Certificate) ValidateTime(comp time.Time) error {
+func (cert *RPKICertificate) ValidateTime(comp time.Time) error {
 	if cert.Certificate == nil {
 		return errors.New("No certificate found")
 	}
@@ -471,7 +543,7 @@ func (cert *RPKI_Certificate) ValidateTime(comp time.Time) error {
 	return nil
 }
 
-func (cert *RPKI_Certificate) String() string {
+func (cert *RPKICertificate) String() string {
 	s := "RPKI Certificate: "
 
 	s += fmt.Sprintf("KeyIdentifier: %v / Emitter: %v",
@@ -544,38 +616,278 @@ func DecodeKeyIdentifier(data []byte) ([]byte, error) {
 	return key, nil
 }
 
-func DecodeCertificate(data []byte) (*RPKI_Certificate, error) {
+// Put in ExtraExtensions
+// https://tools.ietf.org/html/rfc3779
+func GroupIPAddressBlock(ips []IPCertificateInformation) map[byte][]IPCertificateInformation {
+	mapIps := make(map[byte][]IPCertificateInformation)
+	for _, ip := range ips {
+		afi := ip.GetAfi()
+		ipsList, ok := mapIps[afi]
+		if !ok {
+			ipsList = make([]IPCertificateInformation, 0)
+		}
+		ipsList = append(ipsList, ip)
+
+		mapIps[afi] = ipsList
+	}
+	return mapIps
+}
+
+func EncodeInfoAccess(authority bool, path string) (*pkix.Extension, error) {
+	type SubStruct struct {
+		OID  asn1.ObjectIdentifier
+		Path string `asn1:"implicit,tag:6"`
+	}
+
+	oid1 := SubjectInfoAccess
+	oid2 := SignedObject
+	if authority {
+		oid1 = AuthorityInfoAccess
+		oid2 = CAIssuer
+	}
+
+	substruct := SubStruct{
+		OID:  oid2,
+		Path: path,
+	}
+
+	iasBytes, err := asn1.Marshal([]interface{}{substruct})
+	if err != nil {
+		return nil, err
+	}
+	ext := &pkix.Extension{
+		Id:    oid1,
+		Value: iasBytes,
+	}
+	return ext, nil
+}
+
+// https://tools.ietf.org/html/rfc7318
+func EncodePolicyInformation(cps string) (*pkix.Extension, error) {
+	type CertificatePolicy struct {
+		OID    asn1.ObjectIdentifier
+		Policy []interface{}
+	}
+	type CPSStruct struct {
+		OID asn1.ObjectIdentifier
+		CPS string `asn1:"ia5"`
+	}
+	certPolicy := CertificatePolicy{
+		OID: ResourceCertPolicy,
+	}
+	if cps != "" {
+		cpss := CPSStruct{
+			OID: CPS,
+			CPS: cps,
+		}
+
+		certPolicy.Policy = []interface{}{
+			cpss,
+		}
+	}
+
+	policyGroupBytes, err := asn1.Marshal([]CertificatePolicy{certPolicy})
+	if err != nil {
+		return nil, err
+	}
+	ext := &pkix.Extension{
+		Id:       CertPolicy,
+		Critical: true,
+		Value:    policyGroupBytes,
+	}
+	return ext, nil
+}
+
+func EncodeIPAddressBlock(ips []IPCertificateInformation) (*pkix.Extension, error) {
+	groups := GroupIPAddressBlock(ips)
+
+	versionList := make([]int, 0)
+	for version, _ := range groups {
+		versionList = append(versionList, int(version))
+	}
+	sort.Ints(versionList)
+
+	groupAsn1 := make([]asn1.RawValue, 0)
+	for _, cversion := range versionList {
+		version := byte(cversion)
+		ipBytes, err := EncodeIPAddressBlockVersion(version, groups[version], 0, false)
+		if err != nil {
+			return nil, err
+		}
+		groupAsn1 = append(groupAsn1, asn1.RawValue{FullBytes: ipBytes})
+	}
+
+	ipGroupBytes, err := asn1.Marshal(groupAsn1)
+	if err != nil {
+		return nil, err
+	}
+	ext := &pkix.Extension{
+		Id:       IpAddrBlock,
+		Critical: true,
+		Value:    ipGroupBytes,
+	}
+	return ext, nil
+}
+
+func EncodeIPAddressBlockVersion(version byte, ips []IPCertificateInformation, safi byte, addSafi bool) ([]byte, error) {
+	type Ip struct {
+		Version []byte
+		Ips     []asn1.RawValue
+	}
+	type IpNull struct {
+		Version []byte
+		Ips     asn1.RawValue
+	}
+
+	ver := []byte{0, version}
+	if addSafi {
+		ver = append(ver, safi)
+	}
+
+	ipSeq := Ip{
+		Version: ver,
+		Ips:     make([]asn1.RawValue, 0),
+	}
+	for _, ip := range ips {
+		ipBytes, err := ip.ASN1()
+		if err != nil {
+			return nil, err
+		}
+
+		switch ip.(type) {
+		case *IPAddressNull:
+			return asn1.Marshal(IpNull{
+				Version: ver,
+				Ips:     asn1.RawValue{FullBytes: ipBytes},
+			})
+		}
+
+		ipSeq.Ips = append(ipSeq.Ips, asn1.RawValue{FullBytes: ipBytes})
+	}
+
+	return asn1.Marshal(ipSeq)
+}
+
+// https://tools.ietf.org/html/rfc6487
+func EncodeASNSeq(asns []ASNCertificateInformation) ([]asn1.RawValue, error) {
+	if len(asns) == 0 {
+		return nil, nil
+	}
+
+	asnSeq := make([]asn1.RawValue, 0)
+	for _, asn := range asns {
+		asnBytes, err := asn.ASN1()
+		if err != nil {
+			return nil, err
+		}
+		asnStruct := asn1.RawValue{
+			FullBytes: asnBytes,
+		}
+		asnSeq = append(asnSeq, asnStruct)
+
+		switch asn.(type) {
+		case *ASNull:
+			return asnSeq, nil
+		}
+	}
+
+	m, err := asn1.Marshal(asnSeq)
+	if err != nil {
+		return nil, err
+	}
+
+	return []asn1.RawValue{asn1.RawValue{FullBytes: m}}, nil
+}
+
+func EncodeASN(nums []ASNCertificateInformation, rdi []ASNCertificateInformation) (*pkix.Extension, error) {
+	asnSeq, err := EncodeASNSeq(nums)
+	if err != nil {
+		return nil, err
+	}
+	rdiSeq, err := EncodeASNSeq(rdi)
+	if err != nil {
+		return nil, err
+	}
+
+	type AsnStruct struct {
+		ASN []asn1.RawValue `asn1:"tag:0,omitempty"`
+		RDI []asn1.RawValue `asn1:"tag:1,omitempty"`
+	}
+	asnTotalStruct := AsnStruct{
+		ASN: asnSeq,
+		RDI: rdiSeq,
+	}
+
+	asnSeqBytes, err := asn1.Marshal(asnTotalStruct)
+	if err != nil {
+		return nil, err
+	}
+
+	ext := &pkix.Extension{
+		Id:       AutonomousSysIds,
+		Critical: true,
+		Value:    asnSeqBytes,
+	}
+	return ext, nil
+}
+
+func EncodeSIA(sias []*SIA) (*pkix.Extension, error) {
+	siaSeq := make([]asn1.RawValue, 0)
+	for _, sia := range sias {
+		siaBytes, err := asn1.Marshal(*sia)
+		if err != nil {
+			return nil, err
+		}
+		siaStruct := asn1.RawValue{
+			FullBytes: siaBytes,
+		}
+		siaSeq = append(siaSeq, siaStruct)
+	}
+
+	siaSeqBytes, err := asn1.Marshal(siaSeq)
+	if err != nil {
+		return nil, err
+	}
+
+	ext := &pkix.Extension{
+		Id:    SubjectInfoAccess,
+		Value: siaSeqBytes,
+	}
+	return ext, nil
+}
+
+func DecodeCertificate(data []byte) (*RPKICertificate, error) {
 	cert, err := x509.ParseCertificate(data)
 
 	if err != nil {
 		fmt.Print(err)
 		return nil, err
 	}
-	rpki_cert := RPKI_Certificate{
+	rpkiCert := RPKICertificate{
 		Certificate: cert,
 	}
 	for _, extension := range cert.Extensions {
 		if extension.Id.Equal(IpAddrBlock) {
 			addresses, err := DecodeIPAddressBlock(extension.Value)
-			rpki_cert.IPAddresses = addresses
+			rpkiCert.IPAddresses = addresses
 			if err != nil {
-				return &rpki_cert, err
+				return &rpkiCert, err
 			}
 		} else if extension.Id.Equal(AutonomousSysIds) {
 			asnsnum, asnsrdi, err := DecodeASN(extension.Value)
-			rpki_cert.ASNums = asnsnum
-			rpki_cert.ASNRDI = asnsrdi
+			rpkiCert.ASNums = asnsnum
+			rpkiCert.ASNRDI = asnsrdi
 			if err != nil {
-				return &rpki_cert, err
+				return &rpkiCert, err
 			}
 		} else if extension.Id.Equal(SubjectInfoAccess) {
 			sias, err := DecodeSubjectInformationAccess(extension.Value)
-			rpki_cert.SubjectInformationAccess = sias
+			rpkiCert.SubjectInformationAccess = sias
 			if err != nil {
-				return &rpki_cert, err
+				return &rpkiCert, err
 			}
 		}
 	}
 
-	return &rpki_cert, nil
+	return &rpkiCert, nil
 }
